// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "nvfs.h"

static void nvfs_replay_journal(struct nvfs_superblock *nvs, struct nvfs_journal *j)
{
	unsigned i, k;
	bool need_wmb = false;
	for (i = 0; i < NVFS_JOURNAL_LENGTH + 1; i++) {
		uint64_t addr = le64_to_cpu(j->entries[i].address);
		uint64_t val = le64_to_cpu(j->entries[i].value);
		if (!(addr & val & NVFS_JOURNAL_VALID))
			goto not_committed;
		if (addr == NVFS_JOURNAL_END)
			goto committed;
	}
	nvfs_error(nvs, "journal is valid but it doesn't have an end marker");
	goto not_committed;

committed:
	for (k = 0; k < i; k++) {
		uint64_t addr = le64_to_cpu(j->entries[k].address);
		uint64_t val = le64_to_cpu(j->entries[k].value);
		bool size_4 = !!(addr & NVFS_JOURNAL_ADDR_SIZE4);

		val &= ~NVFS_JOURNAL_VALID;
		val |= (addr & NVFS_JOURNAL_ADDR_TOP_VALUE_BIT) * (NVFS_JOURNAL_VALID / NVFS_JOURNAL_ADDR_TOP_VALUE_BIT);
		addr &= ~(NVFS_JOURNAL_VALID | NVFS_JOURNAL_ADDR_TOP_VALUE_BIT | NVFS_JOURNAL_ADDR_SIZE4);

		if (size_4)
			nv_store(nvs, *(__le32 *)(nvs->memory_map + addr), cpu_to_le32(val));
		else
			nv_store(nvs, *(__le64 *)(nvs->memory_map + addr), cpu_to_le64(val));
	}
	nv_wmb(nvs);

not_committed:
	for (i = 0; i < NVFS_JOURNAL_LENGTH + 1; i++) {
		if (le64_to_cpu(j->entries[i].address) | le64_to_cpu(j->entries[i].value)) {
			nv_store(nvs, j->entries[i].address, cpu_to_le64(0));
			nv_store(nvs, j->entries[i].value, cpu_to_le64(0));
			need_wmb = 1;
		}
	}
	if (need_wmb)
		nv_wmb(nvs);
}

static inline void nvfs_replay_quick(struct nvfs_superblock *nvs, struct nvfs_per_cpu *p)
{
	struct nvfs_journal *j;
	unsigned i;
	unsigned n = p->j_idx;
	for (i = 0; i < n; i++) {
		uint64_t addr = p->j_data[i].addr;
		uint64_t val = p->j_data[i].value;
		bool size_4 = !!(addr & NVFS_JOURNAL_ADDR_SIZE4);
		addr &= ~NVFS_JOURNAL_ADDR_SIZE4;

		if (size_4)
			nv_store(nvs, *(__le32 *)(nvs->memory_map + addr), cpu_to_le32(val));
		else
			nv_store(nvs, *(__le64 *)(nvs->memory_map + addr), cpu_to_le64(val));
	}
	nv_wmb(nvs);
	j = p->j;
	for (i = 0; i < n + 1; i++) {
		nv_store(nvs, j->entries[i].address, cpu_to_le64(0));
		nv_store(nvs, j->entries[i].value, cpu_to_le64(0));
	}
	nv_wmb(nvs);
}

void nvfs_start_tx(struct nvfs_superblock *nvs)
{
	struct nvfs_per_cpu *p;
	if (unlikely(nvs->journal_global != NULL)) {
		spin_lock(&nvs->journal_lock);
		p = nvs->journal_global;
	} else {
		preempt_disable();
		p = this_cpu_ptr(nvs->npc);
	}
	BUG_ON(current->journal_info);
	current->journal_info = p;
	p->j_idx = 0;
}

void nvfs_add_to_tx(struct nvfs_superblock *nvs, void *addr, uint64_t val, bool size_4)
{
	uint64_t a = addr - nvs->memory_map;
	struct nvfs_journal_entry *je;
	struct nvfs_per_cpu *p = current->journal_info;
	unsigned idx;

	if (!p) {
		if (!size_4)
			nv_store(nvs, *(__le64 *)addr, cpu_to_le64(val));
		else
			nv_store(nvs, *(__le32 *)addr, cpu_to_le32(val));
		return;
	}

	idx = p->j_idx++;
	BUG_ON(idx >= NVFS_JOURNAL_LENGTH);
	je = &p->j->entries[idx];
	nv_store(nvs, je->address, cpu_to_le64(a
		| NVFS_JOURNAL_VALID
		| ((val & NVFS_JOURNAL_VALID) / (NVFS_JOURNAL_VALID / NVFS_JOURNAL_ADDR_TOP_VALUE_BIT))
		| (size_4 * NVFS_JOURNAL_ADDR_SIZE4)));
	nv_store(nvs, je->value, cpu_to_le64(val | NVFS_JOURNAL_VALID));

	p->j_data[idx].addr = a | (size_4 * NVFS_JOURNAL_ADDR_SIZE4);
	p->j_data[idx].value = val;
}

void nvfs_commit_tx(struct nvfs_superblock *nvs)
{
	struct nvfs_journal_entry *je;
	struct nvfs_per_cpu *p = current->journal_info;

	je = &p->j->entries[p->j_idx];
	nv_store(nvs, je->address, cpu_to_le64(NVFS_JOURNAL_END));
	nv_store(nvs, je->value, cpu_to_le64(NVFS_JOURNAL_END));
	nv_wmb(nvs);

#if 0
	/* test this */
	nvfs_replay_journal(nvs, p->j);
#else
	nvfs_replay_quick(nvs, p);
#endif

	if (unlikely(nvs->journal_global != NULL))
		spin_unlock(&nvs->journal_lock);
	else
		preempt_enable();

	current->journal_info = NULL;
}

__cold int nvfs_init_journal(struct nvfs_superblock *nvs)
{
	unsigned block_size = 1U << nvs->log2_page_size;
	unsigned inline_journals = (block_size - offsetof(struct nvfs_sb, inline_journal.j)) / sizeof(struct nvfs_journal);
	unsigned per_block_journals = (block_size - offsetof(struct nvfs_journal_block, j)) / sizeof(struct nvfs_journal);

	unsigned num_cpus, cpu_idx;
	unsigned num;
	struct nvfs_journal_block *jn;
	struct nvfs_per_cpu *jg;

	spin_lock_init(&nvs->journal_lock);

	//printk("j offset: %u, nvfs_journal: %u\n", (unsigned)(offsetof(struct nvfs_sb, inline_journal.j)), (unsigned)sizeof(struct nvfs_journal));
	//printk("inline_journals: %u, per_block_journals: %u\n", inline_journals, per_block_journals);

	num_cpus = cpumask_last(cpu_possible_mask) + 1;
	cpu_idx = 0;
	jn = &p_superblock(nvs)->inline_journal;
	num = inline_journals;

	while (1) {
		unsigned i;
		for (i = 0; i < num; i++, cpu_idx++) {
			if (cpu_possible(cpu_idx)) {
				struct nvfs_per_cpu *npc = per_cpu_ptr(nvs->npc, cpu_idx);
				npc->j = &jn->j[i];
				npc->j_idx = 0;
			}
			nvfs_replay_journal(nvs, &jn->j[i]);
		}
		if (!le64_to_cpu(jn->next)) {
			uint64_t new_block;
			if (cpu_idx >= num_cpus)
				break;

			if (sb_rdonly(nvs->s))
				goto no_space;
			new_block = nvfs_new_block(nvs, NULL);
			if (!new_block) {
				goto no_space;
			}
			nvfs_memzero_flushcache(nvs, nvs->memory_map + new_block, block_size);
			nv_wmb(nvs);
			jn->next = cpu_to_le64(new_block);
		}
		jn = nvs->memory_map + le64_to_cpu(jn->next);
		num = per_block_journals;
	}

	return 0;

no_space:
	jg = kmalloc(sizeof(struct nvfs_per_cpu), GFP_KERNEL);
	if (!jg)
		return -ENOMEM;
	jg->j = &p_superblock(nvs)->inline_journal.j[0];

	nvs->journal_global = jg;

	return 0;
}

__cold void nvfs_free_journal(struct nvfs_superblock *nvs)
{
	kfree(nvs->journal_global);
}
