// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "nvfs.h"

int nvfs_create(struct user_namespace *ns, struct inode *dir, struct dentry *dentry, umode_t mode, bool excl)
{
	struct inode *inode;
	int r;

	r = dquot_initialize(dir);
	if (unlikely(r))
		goto ret1;

	inode = nvfs_new_inode(ns, dir, mode, 0, &dentry->d_name);

	if (IS_ERR(inode)) {
		r = PTR_ERR(inode);
		goto ret1;
	}

	r = nvfs_add_link(dentry, inode);
	if (unlikely(r))
		goto ret2;

	d_instantiate_new(dentry, inode);
	return 0;

ret2:
	nvfs_dec_link_count(inode);
	discard_new_inode(inode);
ret1:
	return r;
}

int nvfs_mkdir(struct user_namespace *ns, struct inode *dir, struct dentry *dentry, umode_t mode)
{
	struct nvfs_superblock *nvs = dir->i_sb->s_fs_info;
	struct inode *inode;
	int r;

	r = dquot_initialize(dir);
	if (unlikely(r))
		goto ret1;

	nvfs_inc_link_count(dir);

	inode = nvfs_new_inode(ns, dir, S_IFDIR | mode, 0, &dentry->d_name);

	if (IS_ERR(inode)) {
		r = PTR_ERR(inode);
		goto ret2;
	}

	r = nvfs_init_empty_dir(dir, inode);
	if (unlikely(r))
		goto ret3;

	nvfs_inc_link_count(inode);

	nv_wmb(nvs);

	r = nvfs_add_link(dentry, inode);
	if (unlikely(r))
		goto ret4;

	d_instantiate_new(dentry, inode);
	return 0;

ret4:
	nvfs_dec_link_count(inode);
ret3:
	nvfs_dec_link_count(inode);
	discard_new_inode(inode);
ret2:
	nvfs_dec_link_count(dir);
ret1:
	return r;
}

static void nvfs_symlink_create_page(struct nvfs_superblock *nvs, void *page, unsigned prealloc, void *aux)
{
	size_t sl = strlen(aux);
	nv_memcpy_flushcache(nvs, page, aux, sl + 1);
}

int nvfs_symlink(struct user_namespace *ns, struct inode *dir, struct dentry *dentry, const char *symname)
{
	struct nvfs_superblock *nvs = dir->i_sb->s_fs_info;
	size_t sl = strlen(symname);
	struct inode *inode;
	void *m;
	int r;

	if (sl >= 1UL << nvs->log2_page_size) {
		r = -ENAMETOOLONG;
		goto ret1;
	}

	r = dquot_initialize(dir);
	if (unlikely(r))
		goto ret1;

	inode = nvfs_new_inode(ns, dir, S_IFLNK | S_IRWXUGO, 0, &dentry->d_name);
	if (IS_ERR(inode)) {
		r = PTR_ERR(inode);
		goto ret1;
	}

	if (likely(sl < NVFS_FAST_SYMLINK_LEN)) {
		struct nvfs_inode *nvi = i_to_nmi(inode)->nvi;
		nv_memcpy_flushcache(nvs, nvi->i_fast_symlink, symname, sl + 1);
		inode->i_link = nvi->i_fast_symlink;
		i_to_nmi(inode)->i_flag |= NVFS_FLAG_FAST_SYMLINK;
		nv_store(nvs, nvi->i_flag, cpu_to_le32(i_to_nmi(inode)->i_flag));
	}  else {
		m = nvfs_bmap(i_to_nmi(inode), 0, NULL, nvfs_symlink_create_page, (void *)symname, NULL);
		if (unlikely(IS_ERR(m))) {
			r = PTR_ERR(m);
			goto ret2;
		}
		inode->i_link = m;
	}
	nvfs_set_inode_size(i_to_nmi(inode), sl);

	r = nvfs_add_link(dentry, inode);
	if (unlikely(r))
		goto ret2;

	d_instantiate_new(dentry, inode);
	return 0;

ret2:
	nvfs_dec_link_count(inode);
	discard_new_inode(inode);
ret1:
	return r;
}

int nvfs_mknod(struct user_namespace *ns, struct inode *dir, struct dentry *dentry, umode_t mode, dev_t rdev)
{
	struct inode *inode;
	int r;

	r = dquot_initialize(dir);
	if (unlikely(r))
		goto ret1;

	inode = nvfs_new_inode(ns, dir, mode, rdev, &dentry->d_name);
	if (IS_ERR(inode)) {
		r = PTR_ERR(inode);
		goto ret1;
	}

	r = nvfs_add_link(dentry, inode);
	if (unlikely(r))
		goto ret2;

	d_instantiate_new(dentry, inode);
	return 0;

ret2:
	nvfs_dec_link_count(inode);
	discard_new_inode(inode);
ret1:
	return r;
}

int nvfs_unlink(struct inode *dir, struct dentry *dentry)
{
	struct nvfs_superblock *nvs;
	int r;
	struct inode *inode;
	struct nvfs_dir_entry *nde;

	r = dquot_initialize(dir);
	if (unlikely(r))
		return r;

	nde = nvfs_prepare_delete(i_to_nmi(dir), &dentry->d_name);
	if (IS_ERR(nde))
		return PTR_ERR(nde);

	nvs = dir->i_sb->s_fs_info;
	nvfs_finish_delete(nvs, nde);

	nv_wmb(nvs);

	inode = d_inode(dentry);
	nvfs_update_time(inode, &dir->i_ctime, S_CTIME);

	nvfs_dec_link_count(inode);
	return 0;
}

int nvfs_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct inode *inode = d_inode(dentry);
	int r;

	r = nvfs_is_dir_empty(i_to_nmi(inode));
	if (unlikely(r < 0))
		return r;
	if (unlikely(!r))
		return -ENOTEMPTY;

	r = nvfs_unlink(dir, dentry);
	if (unlikely(r < 0))
		return r;

	inode->i_size = 0;
	nvfs_dec_link_count(inode);
	nvfs_dec_link_count(dir);

	return 0;
}

static int nvfs_exchange(struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry)
{
	struct inode *old_inode = d_inode(old_dentry);
	struct inode *new_inode = d_inode(new_dentry);
	struct nvfs_superblock *nvs = old_inode->i_sb->s_fs_info;
	struct nvfs_dir_entry *old_nde, *new_nde;
	uint64_t old_ino, new_ino;
	uint32_t old_flags, new_flags;
	int link_diff;

	nvfs_update_time(old_inode, NULL, S_CTIME);
	nvfs_update_time(new_inode, NULL, S_CTIME);

	old_nde = nvfs_prepare_delete(i_to_nmi(old_dir), &old_dentry->d_name);
	if (IS_ERR(old_nde))
		return PTR_ERR(old_nde);

	new_nde = nvfs_prepare_delete(i_to_nmi(new_dir), &new_dentry->d_name);
	if (IS_ERR(new_nde))
		return PTR_ERR(new_nde);

	old_ino = le64_to_cpu(old_nde->inode);
	new_ino = le64_to_cpu(new_nde->inode);
	old_flags = le32_to_cpu(old_nde->name_len_and_type);
	new_flags = le32_to_cpu(new_nde->name_len_and_type);

	nvfs_start_tx(nvs);

	nvfs_add_to_tx(nvs, &old_nde->inode, new_ino, false);
	nvfs_add_to_tx(nvs, &new_nde->inode, old_ino, false);
	nvfs_add_to_tx(nvs, &old_nde->name_len_and_type, (old_flags & NAME_LEN_MASK) | (new_flags & TYPE_MASK), true);
	nvfs_add_to_tx(nvs, &new_nde->name_len_and_type, (new_flags & NAME_LEN_MASK) | (old_flags & TYPE_MASK), true);

	nvfs_commit_tx(nvs);

	if (S_ISDIR(old_inode->i_mode)) {
		struct nvfs_memory_inode *nmi;
		nmi = i_to_nmi(old_inode);
		nmi->nvi->i_parent = cpu_to_le64(new_dir->i_ino);
	}
	if (S_ISDIR(new_inode->i_mode)) {
		struct nvfs_memory_inode *nmi;
		nmi = i_to_nmi(new_inode);
		nmi->nvi->i_parent = cpu_to_le64(old_dir->i_ino);
	}

	link_diff = !!S_ISDIR(old_inode->i_mode) - !!S_ISDIR(new_inode->i_mode);

	if (unlikely(link_diff < 0)) {
		nvfs_inc_link_count(old_dir);
		nvfs_dec_link_count(new_dir);
	} else if (unlikely(link_diff > 0)) {
		nvfs_inc_link_count(new_dir);
		nvfs_dec_link_count(old_dir);
	}

	return 0;
}

int nvfs_rename(struct user_namespace *ns, struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry, unsigned flags)
{
	struct inode *old_inode = d_inode(old_dentry);
	struct inode *new_inode = d_inode(new_dentry);
	struct nvfs_superblock *nvs = old_inode->i_sb->s_fs_info;
	int r;
	struct nvfs_dir_entry *nde, *del_nde;

	if (unlikely(flags & ~(RENAME_NOREPLACE | RENAME_EXCHANGE)))
		return -EINVAL;

	r = dquot_initialize(old_dir);
	if (unlikely(r))
		return r;

	r = dquot_initialize(new_dir);
	if (unlikely(r))
		return r;

	if (unlikely((flags & RENAME_EXCHANGE) != 0)) {
		return nvfs_exchange(old_dir, old_dentry, new_dir, new_dentry);
	}

	del_nde = nvfs_prepare_delete(i_to_nmi(old_dir), &old_dentry->d_name);
	if (IS_ERR(del_nde))
		return PTR_ERR(del_nde);

	if (new_inode) {
		if (S_ISDIR(new_inode->i_mode)) {
			r = nvfs_is_dir_empty(i_to_nmi(new_inode));
			if (unlikely(r < 0))
				return r;
			if (unlikely(!r))
				return -ENOTEMPTY;
		}

		nvfs_update_time(new_inode, NULL, S_CTIME);

		nde = nvfs_find_de(i_to_nmi(new_dir), &new_dentry->d_name);
		if (unlikely(IS_ERR(nde)))
			return PTR_ERR(nde);
		if (unlikely(!nde)) {
			nvfs_error(nvs, "rename: directory entry for a new file '%s' disappeared", new_dentry->d_name.name);
			return -EUCLEAN;
		}

		nvfs_update_time(new_dir, NULL, S_CTIME | S_MTIME);
		nvfs_update_time(old_inode, NULL, S_CTIME);

		nvfs_start_tx(nvs);

		nvfs_add_to_tx(nvs, &nde->inode, old_inode->i_ino, false);
	} else {
		nde = nvfs_prepare_add(new_dentry, old_inode);
		if (unlikely(IS_ERR(nde)))
			return PTR_ERR(nde);
		if (S_ISDIR(old_inode->i_mode))
			nvfs_inc_link_count(new_dir);

		nvfs_update_time(old_inode, NULL, S_CTIME);

		nvfs_start_tx(nvs);
	}

	nvfs_finish_delete_tx(nvs, del_nde);
	nvfs_finish_add(nde, new_dentry, old_inode);
	nvfs_commit_tx(nvs);

	if (S_ISDIR(old_inode->i_mode)) {
		struct nvfs_memory_inode *nmi;
		nvfs_dec_link_count(old_dir);
		nmi = i_to_nmi(old_inode);
		nmi->nvi->i_parent = cpu_to_le64(new_dir->i_ino);
	}

	if (new_inode) {
		nvfs_dec_link_count(new_inode);
		if (S_ISDIR(new_inode->i_mode)) {
			nvfs_dec_link_count(new_inode);
			new_inode->i_size = 0;
		}
	}

	return 0;
}

int nvfs_link(struct dentry *old_dentry, struct inode *dir, struct dentry *dentry)
{
	struct inode *inode = d_inode(old_dentry);
	int r;

	r = dquot_initialize(dir);
	if (unlikely(r))
		return r;

	nvfs_inc_link_count(inode);
	ihold(inode);

	r = nvfs_add_link(dentry, inode);
	if (unlikely(r)) {
		nvfs_dec_link_count(inode);
		iput(inode);
		return r;
	}

	nvfs_update_time(inode, NULL, S_CTIME);

	d_instantiate(dentry, inode);
	return 0;
}
