// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "fsck.h"

static uint64_t block_bytes;
static _Atomic uint64_t *block_map = NULL;

static uint64_t inode_bytes;
static _Atomic uint64_t *inode_map = NULL;

static uint64_t nlink_bytes;
static _Atomic uint32_t *nlink_map = NULL;

bool is_block_reserved(uint64_t block)
{
	uint64_t grp = block >> log2_pages_per_group;
	uint64_t off = block & ((1 << log2_pages_per_group) - 1);
	if (!grp && !off)
		return true;
	if (off >= 1 && off < 2 + (1U << sb->log2_inode_pages_per_group))
		return true;
	if (block >= n_groups << log2_pages_per_group)
		return true;
	if (block >= le64_to_cpu(sb->n_pages))
		return true;
	return false;
}

bool is_inode_reserved(uint64_t inode)
{
	uint64_t grp = inode >> log2_inodes_per_group;
	unsigned off = inode & ((1 << log2_inodes_per_group) - 1);
	if (off < inode_bitmap_inodes)
		return true;
	if (grp == n_groups - 1) {
		if (off >= inode_bitmap_inodes + inodes_in_last_group)
			return true;
	}
	return false;
}

void *group_ptr(uint64_t group)
{
	return ptr + (group << (sb->log2_page_size + log2_pages_per_group));
}

bool bmp_op(uint64_t b, unsigned flags)
{
	_Atomic uint64_t *p;
	unsigned off;
	uint64_t x, y;

	if (flags & B_BLOCK) {
		if (flags & B_MEM) {
			if (b >= block_bytes * 8)
				fsck_internal_error("bmp_op: invalid block: %"PRIu64" >= %"PRIu64"", b, block_bytes * 8);

			p = &block_map[b / 64];
		} else {
			uint64_t group = b >> log2_pages_per_group;
			unsigned bit = b & ((1 << log2_pages_per_group) - 1);
			void *bmp;
			if (group >= n_groups)
				fsck_internal_error("bmp_op: invalid block group: %"PRIu64" >= %"PRIu64"", group, n_groups);
			bmp = group_ptr(group) + (1 << sb->log2_page_size);
			//fprintf(stderr, "bmp offset: %lx (%lx, %x)\n", distance(bmp), (group << (sb->log2_page_size + log2_pages_per_group)), 1 << sb->log2_page_size);
			p = &((_Atomic uint64_t *)bmp)[bit / 64];
		}
	} else {
		if (flags & B_MEM) {
			if (b >= inode_bytes * 8)
				fsck_internal_error("bmp_op: invalid inode: %"PRIu64" >= %"PRIu64"", b, inode_bytes * 8);

			p = &inode_map[b / 64];
		} else {
			uint64_t group = b >> log2_inodes_per_group;
			unsigned bit = b & ((1 << log2_inodes_per_group) - 1);
			void *bmp;
			if (group >= n_groups)
				fsck_internal_error("bmp_op: invalid inode group: %"PRIu64" >= %"PRIu64"", group, n_groups);
			bmp = group_ptr(group) + (2 << sb->log2_page_size) + offsetof(struct nvfs_group, bitmap);
			p = &((_Atomic uint64_t *)bmp)[bit / 64];
		}
	}
	off = b % 64;
#if __BYTE_ORDER == __BIG_ENDIAN
	off ^= 56;
#endif

	x = atomic_load(p);

	if (flags & B_READ)
		return (x >> off) & 1;

again:
	if (flags & B_SET) {
		if (x & (1ULL << off))
			return false;
		y = x | (1ULL << off);
	} else {
		if (!(x & (1ULL << off)))
			return false;
		y = x & ~(1ULL << off);
	}
	if (!atomic_compare_exchange_weak(p, &x, y))
		goto again;

	return true;
}

bool inc_n_links(uint64_t ino)
{
	uint32_t nl, new_nl;
	if (bmp_op(ino, B_SET | B_INODE | B_MEM))
		return true;
	nl = nlink_map[ino];
again:
	new_nl = nl + 1;
	if (new_nl == 0xFFFFFFFFU)
		return false;
	if (!atomic_compare_exchange_weak(&nlink_map[ino], &nl, new_nl))
		goto again;
	return true;
}

void dec_n_links(uint64_t ino)
{
	nlink_map[ino]--;
}

uint32_t get_n_links(uint64_t ino)
{
	return nlink_map[ino] + 1;
}

bool test_ino_flag(uint64_t ino)
{
	return nlink_map[ino];
}

void set_ino_flag(uint64_t ino)
{
	nlink_map[ino] = 1;
}

void clear_ino_flag(uint64_t ino)
{
	nlink_map[ino] = 0;
}

static atomic_ullong last_block_alloc;
static atomic_ullong last_inode_alloc;

uint64_t alloc_block_or_inode(unsigned flags)
{
	uint64_t b, n, limit;
	// !!! TODO: yse a better algorithm than block-by-block scan.
	if (flags & B_BLOCK) {
		limit = n = n_groups << log2_pages_per_group;
		b = last_block_alloc;
	} else {
		limit = n = n_groups << log2_inodes_per_group;
		b = last_inode_alloc;
	}
	while (n--) {
		if (b >= limit)
			b = 0;
		if (!bmp_op(b, flags | B_READ | B_MEM)) {
			if (bmp_op(b, flags | B_SET | B_MEM)) {
				if (bmp_op(b, flags | B_SET | B_DEV)) {
					if (flags & B_BLOCK)
						last_block_alloc = b;
					else
						last_inode_alloc = b;
					return b;
				} else {
					bmp_op(b, flags | B_CLEAR | B_MEM);
				}
			}
		}
		b++;
	}
	return 0;
}

static void fill_blockmap(void)
{
	uint64_t b;
	//fprintf(stderr, "filling blockmap...\n");
#pragma omp parallel for
	for (b = 0; b < n_groups << log2_pages_per_group; b++) {
		if (is_block_reserved(b)) {
			if (!bmp_op(b, B_SET | B_BLOCK | B_MEM))
				fsck_internal_error("init_blockmap: block: %"PRIu64" already set", b);
			if (!bmp_op(b, B_READ | B_BLOCK | B_DEV)) {
				static struct qkey q;
				if (fsck_query(&q, Q_NORMAL, "reserved block %"PRIu64" is not marked in the bitmap", "fix?", b)) {
					bmp_op(b, B_SET | B_BLOCK | B_DEV);
				}
			}
			//fprintf(stderr, "block: %lx\n", b);
		}
	}
	//fprintf(stderr, "filling blockmap done...\n");
}

static void fill_inodemap(void)
{
	uint64_t b;
	//fprintf(stderr, "filling inodemap...\n");
#pragma omp parallel for
	for (b = 0; b < n_groups << log2_inodes_per_group; b++) {
		if (is_inode_reserved(b)) {
			if (!bmp_op(b, B_SET | B_INODE | B_MEM))
				fsck_internal_error("init_inodemap: inode: %"PRIu64" already set", b);
			if (!bmp_op(b, B_READ | B_INODE | B_DEV)) {
				static struct qkey q;
				if (fsck_query(&q, Q_NORMAL, "reserved inode %"PRIu64" is not marked in the bitmap", "fix?", b)) {
					bmp_op(b, B_SET | B_INODE | B_DEV);
				}
			}
			//fprintf(stderr, "inode: %lx\n", b);
		}
	}
	//fprintf(stderr, "filling inodemap done...\n");
}

int init_blockmap(void)
{
	uint64_t rn;

	rn = random() + (uint64_t)random() * RAND_MAX;
	rn %= n_groups << log2_pages_per_group;
	last_block_alloc = rn;
	rn = random() + (uint64_t)random() * RAND_MAX;
	rn %= n_groups << log2_inodes_per_group;
	last_inode_alloc = rn;

	block_bytes = ((n_groups << log2_pages_per_group) + 7) / 8;
	if (block_bytes != (size_t)block_bytes) {
		fsck_log("too large device for a 32-bit system");
		goto err;
	}
	block_map = fsck_mmap(block_bytes);

	fill_blockmap();

	inode_bytes = ((n_groups << log2_inodes_per_group) + 7) / 8;
	if (inode_bytes != (size_t)inode_bytes) {
		fsck_log("too large device for a 32-bit system");
		goto err;
	}
	inode_map = fsck_mmap(inode_bytes);

	fill_inodemap();

	nlink_bytes = (n_groups << log2_inodes_per_group) * sizeof(_Atomic uint32_t);
	if (nlink_bytes != (size_t)nlink_bytes) {
		fsck_log("too large device for a 32-bit system");
		goto err;
	}
	nlink_map = fsck_mmap(nlink_bytes);

	return 0;

err:
	done_blockmap();
	return 8;
}

void done_blockmap(void)
{
	if (block_map) {
		fsck_munmap(block_map, block_bytes);
		block_map = NULL;
	}
	if (inode_map) {
		fsck_munmap(inode_map, inode_bytes);
		inode_map = NULL;
	}
	if (nlink_map) {
		fsck_munmap(nlink_map, nlink_bytes);
		nlink_map = NULL;
	}
}
