// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "fsck.h"

static unsigned long hash_to_ptr(__le64 *p, unsigned hash, int hash_shift, unsigned bits)
{
	unsigned offs;
	unsigned long block;

	if (hash_shift > 0)
		offs = (hash >> hash_shift);
	else
		offs = (hash << -hash_shift);
	offs &= (1U << bits) - 1;
	block = le64_to_cpu(p[offs]);
	return block;
}

#define NVFS_FIND_CHAIN			-1
#define NVFS_FIND_NEXT_LEVEL		-2

struct split_info {
	__le64 *ptr;
	unsigned bits;
	int split_bit;
};

static struct nvfs_dir_head *nvfs_find_page(struct nvfs_inode *dir, uint32_t hash, uint32_t *next_hash, struct split_info *si)
{
	struct nvfs_dir_head *ndh;
	__le64 *p = dir->i_radix;
	unsigned bits = NVFS_LOG2_RADIX_BLOCKS;
	int hash_shift = 32 - NVFS_LOG2_RADIX_BLOCKS;
	unsigned long block, block2;

again:
	block = hash_to_ptr(p, hash, hash_shift, bits);

	if (!verify_block_range(block))
		return NULL;

	ndh = ptr + block;
	if (ndh->magic == NVFS_DIR_HEAD_MAGIC) {
		if (si != NULL) {
			int i;
			si->ptr = p;
			si->bits = bits;
			si->split_bit = hash_shift > 0 ? NVFS_FIND_NEXT_LEVEL : NVFS_FIND_CHAIN;
			for (i = hash_shift < 0 ? 0 : hash_shift; (unsigned)i < hash_shift + bits; i++) {
				uint32_t hash2 = hash ^ (1U << i);
				block2 = hash_to_ptr(p, hash2, hash_shift, bits);
				if (block2 == block) {
					si->split_bit = i;
				}
			}
		}
		if (next_hash != NULL) {
			int i;
			for (i = hash_shift < 0 ? 0 : hash_shift; (unsigned)i < hash_shift + bits; i++) {
				if (!((hash >> i) & 1)) {
					uint32_t hash2 = hash | (1U << i);
					block2 = hash_to_ptr(ptr, hash2, hash_shift, bits);
					if (block2 == block) {
						hash = hash2;
					} else {
						break;
					}
				}
			}
			if (hash_shift >= 0)
				*next_hash = hash + (1U << hash_shift);
			else
				*next_hash = hash + 1U;
		}
		return ndh;
	}

	p = (__le64 *)ndh;
	bits = sb->log2_page_size - 3;
	hash_shift -= bits;

	goto again;
}

static struct nvfs_dir_entry *nvfs_prepare_page(struct nvfs_dir_head *ndh, const char *name, unsigned namelen)
{
	struct nvfs_dir_entry *nde, *end;
	size_t needed_len;

	nde = ndh->nde;
	end = (void *)ndh + (1U << sb->log2_page_size);
	needed_len = nvfs_dir_entry_needed_size(namelen);
	do {
		size_t de_size = nvfs_dir_entry_size(nde);
		size_t nns;

		if ((size_t)((void *)end - (void *)nde) < de_size) {
			static struct qkey q;
			fsck_query(&q, Q_JUST_PRINT, "inode %"PRIu64" has invalid leaf directory block", NULL, le64_to_cpu(ndh->inode));
			return NULL;
		}

		if (le32_to_cpu(nde->name_len_and_type) & TYPE_MASK)
			goto used;

		if (nvfs_dir_entry_size(nde) < needed_len) {
			size_t next_size;
			struct nvfs_dir_entry *next_nde = (void *)nde + de_size;
			if (next_nde >= end)
				goto used;
			if (le32_to_cpu(next_nde->name_len_and_type) & TYPE_MASK)
				goto used;
			next_size = de_size + nvfs_dir_entry_size(next_nde);
			nde->name_len_and_type = cpu_to_le32(nvfs_dir_entry_make_size(next_size) << NAME_LEN_OFFSET);
			continue;
		}

		nns = nvfs_dir_entry_needed_size(le32_to_cpu(nde->name_len_and_type));
		if (nns > needed_len) {
			size_t remaining;
			struct nvfs_dir_entry *next_nde;
			remaining = nns - needed_len;
			if (remaining < 16)
				goto used;
			next_nde = (void *)nde + needed_len;
			next_nde->name_len_and_type = cpu_to_le32(nvfs_dir_entry_make_size(remaining) << NAME_LEN_OFFSET);
		}

		memset(nde->name, 0, 4);
		memcpy(nde->name, name, namelen);
		nde->name_len_and_type = cpu_to_le32(namelen << NAME_LEN_OFFSET);
		return nde;

used:
		nde = (void *)nde + de_size;
	} while (nde < end);

	return NULL;
}

static void nvfs_split_page(struct nvfs_dir_head *ndh, struct nvfs_dir_head *ndh1, struct nvfs_dir_head *ndh2, unsigned bit)
{
	struct nvfs_dir_entry *nde, *nde1, *nde2;
	struct nvfs_dir_entry *end, *end1, *end2;
	size_t left1, left2;

	nde = ndh->nde;
	nde1 = ndh1->nde;
	nde2 = ndh2->nde;
	end = (void *)ndh + (1UL << sb->log2_page_size);
	do {
		uint32_t hash;
		size_t de_size = nvfs_dir_entry_size(nde);
		if ((size_t)((void *)end - (void *)nde) < de_size) {
			static struct qkey q;
			fsck_query(&q, Q_JUST_PRINT, "inode %"PRIu64" has invalid leaf directory block", NULL, le64_to_cpu(ndh->inode));
			break;
		}
		if (!nvfs_dir_entry_type(nde))
			goto next_entry;
		hash = nvfs_hash_name(nde->name, nvfs_dir_entry_namelen(nde));
		if (!((hash >> bit) & 1)) {
			memcpy(nde1, nde, de_size);
			nde1 = (void *)nde1 + de_size;
		} else {
			memcpy(nde2, nde, de_size);
			nde2 = (void *)nde2 + de_size;
		}
next_entry:
		nde = (void *)nde + de_size;
	} while (nde < end);

	end1 = (void *)ndh1 + (1UL << sb->log2_page_size);
	end2 = (void *)ndh2 + (1UL << sb->log2_page_size);

	left1 = (void *)end1 - (void *)nde1;
	left2 = (void *)end2 - (void *)nde2;
	if (left1)
		nde1->name_len_and_type = cpu_to_le32(nvfs_dir_entry_make_size(left1));
	if (left2)
		nde2->name_len_and_type = cpu_to_le32(nvfs_dir_entry_make_size(left2));
}

struct nvfs_dir_entry *nvfs_find_de(uint64_t dir_ino, const char *name, unsigned namelen)
{
	uint32_t hash = nvfs_hash_name((uint8_t *)name, namelen);
	struct nvfs_dir_head *ndh;
	struct nvfs_dir_entry *nde, *end;
	uint64_t next;
	__le32 key;

	if (namelen >= 4) {
		memcpy(&key, name, 4);
	} else {
		key = 0;
		memcpy(&key, name, namelen);
	}

	ndh = nvfs_find_page(p_inode(dir_ino), hash, NULL, NULL);
	if (!ndh)
		return NULL;
next_page:
	end = (void *)ndh + (1U << sb->log2_page_size);
	nde = ndh->nde;
	do {
		size_t de_size = nvfs_dir_entry_size(nde);
		if ((size_t)((void *)end - (void *)nde) < de_size) {
			static struct qkey q;
			fsck_query(&q, Q_JUST_PRINT, "inode %"PRIu64" has invalid leaf directory block", NULL, dir_ino);
			return NULL;
		}
		if (*(__le32 *)nde->name != key)
			goto next_entry;
		if (!nvfs_dir_entry_type(nde))
			goto next_entry;
		if (namelen != nvfs_dir_entry_namelen(nde))
			goto next_entry;
		if (memcmp(nde->name, name, namelen))
			goto next_entry;
		return nde;
next_entry:
		nde = (void *)nde + nvfs_dir_entry_size(nde);
	} while (nde < end);
	next = le64_to_cpu(ndh->next);
	if (next != 0) {
		if (!verify_block_range(next))
			return NULL;
		ndh = ptr + next;
		goto next_page;
	}

	return NULL;
}

static void inc_dir_size(struct nvfs_inode *dir_inode)
{
	dir_inode->i_size = cpu_to_le64(le64_to_cpu(dir_inode->i_size) + (1U << sb->log2_page_size));
}

struct nvfs_dir_entry *add_directory_entry(uint64_t dir_ino, const char *name, unsigned namelen, uint64_t target_ino, unsigned char dtype)
{
	uint32_t hash = nvfs_hash_name((uint8_t *)name, namelen);
	struct nvfs_inode *dir_inode = p_inode(dir_ino);
	struct nvfs_dir_head *ndh;
	struct nvfs_dir_entry *nde;
	unsigned long n;
	struct split_info si;
	bool need_chain;

again:
	need_chain = false;
	ndh = nvfs_find_page(dir_inode, hash, NULL, &si);
	if (!ndh)
		return NULL;

	n = 0;
next_page:
	nde = nvfs_prepare_page(ndh, name, namelen);
	if (nde) {
		nde->inode = cpu_to_le64(target_ino);
		nde->name_len_and_type |= cpu_to_le32(dtype << TYPE_OFFSET);
		return nde;
	}

	if (si.split_bit == NVFS_FIND_CHAIN || le64_to_cpu(ndh->next) || need_chain) {
		uint64_t page_n, new_page;

		need_chain = true;

		page_n = le64_to_cpu(ndh->next);
		if (page_n != 0) {
			if (!verify_block_range(page_n))
				return NULL;
			ndh = ptr + page_n;
			n++;
			goto next_page;
		}

		if (n >= nvfs_max_chain_length(sb->log2_page_size))
			return NULL;

		new_page = alloc_dir_leaf(dir_ino);
		if (!new_page)
			return NULL;

		inc_dir_size(dir_inode);

		ndh->next = cpu_to_le64(new_page);

		ndh = ptr + new_page;
		goto next_page;
	} else if (si.split_bit == NVFS_FIND_NEXT_LEVEL) {
		unsigned i;
		uint64_t new_block, block;
		__le64 *new_page;

		new_block = alloc_block_or_inode(B_BLOCK);
		if (!new_block)
			return NULL;

		inc_dir_size(dir_inode);

		new_page = ptr + (new_block << sb->log2_page_size);
		block = distance(ndh);

		for (i = 0; i < 1U << (sb->log2_page_size - 3); i++) {
			new_page[i] = cpu_to_le64(block);
		}

		for (i = 0; i < 1U << si.bits; i++) {
			if (le64_to_cpu(si.ptr[i]) == block) {
				si.ptr[i] = cpu_to_le64(distance(new_page));
				goto again;
			}
		}
		fsck_internal_error("add_directory_entry: corrupted directory tree when creating new level");
	} else {
		uint64_t b1, b2, block;
		struct nvfs_dir_head *ndh1, *ndh2;
		unsigned i, j, k;

		b1 = alloc_dir_leaf(dir_ino);
		if (!b1)
			return NULL;
		b2 = alloc_dir_leaf(dir_ino);
		if (!b2) {
			bmp_op(b1 >> sb->log2_page_size, B_BLOCK | B_CLEAR | B_MEM);
			bmp_op(b1 >> sb->log2_page_size, B_BLOCK | B_CLEAR | B_DEV);
			return NULL;
		}

		ndh1 = ptr + b1;
		ndh2 = ptr + b2;

		nvfs_split_page(ndh, ndh1, ndh2, si.split_bit);

		block = distance(ndh);
		for (i = 0; i < 1U << si.bits; i++) {
			if (le64_to_cpu(si.ptr[i]) == block) {
				goto found2;
			}
		}
		fsck_internal_error("add_directory_entry: corrupted directory tree when splitting page");

found2:
		for (j = i; j < 1U << si.bits; j++) {
			if (le64_to_cpu(si.ptr[j]) != block) {
				break;
			}
		}

		for (k = i; k < j; k++) {
			si.ptr[k] = cpu_to_le64(k < (i + j) / 2 ? b1 : b2);
		}

		bmp_op(distance(ndh) >> sb->log2_page_size, B_BLOCK | B_CLEAR | B_MEM);
		bmp_op(distance(ndh) >> sb->log2_page_size, B_BLOCK | B_CLEAR | B_DEV);

		inc_dir_size(dir_inode);

		goto again;
	}
}
