// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "fsck.h"

static void check_j(struct nvfs_journal *j, unsigned idx)
{
	unsigned i, k;
	const char *reason;
	for (i = 0; i < NVFS_JOURNAL_LENGTH + 1; i++) {
		uint64_t addr = le64_to_cpu(j->entries[i].address);
		uint64_t val = le64_to_cpu(j->entries[i].value);
		if (!(addr & val & NVFS_JOURNAL_VALID))
			goto not_committed;
		if (addr == NVFS_JOURNAL_END)
			goto committed;
	}
	reason = "journal is valid but it doesn't have an end marker";
	goto j_error;

committed:
	for (k = 0; k < i; k++) {
		uint64_t addr = le64_to_cpu(j->entries[k].address);
		unsigned sz = addr & NVFS_JOURNAL_ADDR_SIZE4 ? 4 : 8;
		addr &= ~(NVFS_JOURNAL_VALID | NVFS_JOURNAL_ADDR_TOP_VALUE_BIT | NVFS_JOURNAL_ADDR_SIZE4);
		if (addr + sz - 1 >= le64_to_cpu(sb->n_pages) << sb->log2_page_size) {
			reason = "pointer beyond device end";
			goto j_error;
		}
		if (addr & (sz - 1)) {
			reason = "unaligned address";
			goto j_error;
		}
	}

	{
		static struct qkey q;
		if (fsck_query(&q, Q_PREEN, "journal %u is valid", "replay?", idx)) {
			for (k = 0; k < i; k++) {
				uint64_t addr = le64_to_cpu(j->entries[k].address);
				uint64_t val = le64_to_cpu(j->entries[k].value);
				bool size_4 = !!(addr & NVFS_JOURNAL_ADDR_SIZE4);
				val &= ~NVFS_JOURNAL_VALID;
				val |= (addr & NVFS_JOURNAL_ADDR_TOP_VALUE_BIT) * (NVFS_JOURNAL_VALID / NVFS_JOURNAL_ADDR_TOP_VALUE_BIT);
				addr &= ~(NVFS_JOURNAL_VALID | NVFS_JOURNAL_ADDR_TOP_VALUE_BIT | NVFS_JOURNAL_ADDR_SIZE4);
				if (size_4) {
					*(uint32_t *)(ptr + addr) = cpu_to_le32(val);
				} else {
					*(uint64_t *)(ptr + addr) = cpu_to_le64(val);
				}
			}
			goto clear_and_exit;
		}
	}

not_committed:
	for (k = 0; k < NVFS_JOURNAL_LENGTH + 1; k++) {
		uint64_t addr = le64_to_cpu(j->entries[k].address);
		uint64_t val = le64_to_cpu(j->entries[k].value);
		if (addr | val) {
			static struct qkey q;
			if (fsck_query(&q, Q_PREEN, "journal %u is incomplete", "clear?", idx)) {
				goto clear_and_exit;
			}
			break;
		}
	}
	return;

j_error:
	{
		static struct qkey q;
		if (fsck_query(&q, Q_NORMAL, "journal %u is invalid (%s)", "clear?", idx, reason)) {
clear_and_exit:
			memset(j, 0, sizeof(struct nvfs_journal));
		}
	}
}

int check_journal(void)
{
	unsigned block_size = 1U << sb->log2_page_size;
	unsigned inline_journals = (block_size - offsetof(struct nvfs_sb, inline_journal.j)) / sizeof(struct nvfs_journal);
	unsigned per_block_journals = (block_size - offsetof(struct nvfs_journal_block, j)) / sizeof(struct nvfs_journal);
	unsigned num, idx;
	struct nvfs_journal_block *jn;

	num = inline_journals;
	idx = 0;
	jn = &sb->inline_journal;

	while (1) {
		unsigned i;
		for (i = 0; i < num; i++, idx++) {
			struct nvfs_journal *j = &jn->j[i];
			check_j(j, idx);
		}
		if (!le64_to_cpu(jn->next))
			break;
		if (!verify_block(0, &jn->next, false)) {
			static struct qkey q;
			if (fsck_query(&q, Q_NORMAL, "invalid journal chain", "clear?")) {
				jn->next = cpu_to_le64(0);
				return 0;
			}
		}
		jn = ptr + le64_to_cpu(jn->next);
		num = per_block_journals;
	}

	return 0;
}
