// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

struct test {
	const char *name;
	void (*prepare)(void);
	void (*test)(void);
	void (*check)(void);
};

static void xsystem(const char *cmd)
{
	int r;
	r = system(cmd);
	if (r) fprintf(stderr, "error '%s': %d\n", cmd, r), exit(1);
}

static void do_test(const struct test *t)
{
	int r;
	fprintf(stderr, "================== %s ==================\n", t->name);
	//xsystem("dd if=/dev/zero of=/dev/pmem0 bs=1M oflag=direct count=16");
	xsystem("./fuzz /dev/pmem0");
	xsystem("./mknvfs /dev/pmem0");
	xsystem("mount -t nvfs /dev/pmem0 /mnt/test");
	t->prepare();
	xsystem("umount /mnt/test");
	xsystem("cp /dev/pmem0 /tmp/pmem0.backup");
	xsystem("mount -t nvfs /dev/pmem0 -o log=log.bin /mnt/test");
	t->test();
	xsystem("umount /mnt/test");
	xsystem("cp /dev/pmem0 /tmp/pmem0.backup2");
	xsystem("cp /tmp/pmem0.backup /dev/pmem0");
	xsystem("./replay /dev/pmem0 log.bin");
	r = system("./nvfsck -a /dev/pmem0");
	if (r == -1) perror("./nvfsck -a /dev/pmem0"), exit(1);
	if (WEXITSTATUS(r) >= 4) fprintf(stderr, "fsck failed with error code %d, %d\n", WEXITSTATUS(r), r), exit(1);
	xsystem("./nvfsck -fn /dev/pmem0");
	xsystem("mount -t nvfs /dev/pmem0 /mnt/test");
	t->check();
	xsystem("umount /mnt/test");
}


static void noop(void)
{
}


static void test_create(void)
{
	xsystem("echo Hello > /mnt/test/a.txt");
	xsystem("rm /mnt/test/a.txt");
}

static void check_create(void)
{
	struct stat st;
	if (stat("/mnt/test/a.txt", &st) == -1) {
		if (errno == ENOENT)
			return;
		perror("stat"), exit(1);
	}
	if (!S_ISREG(st.st_mode))
		fprintf(stderr, "bad mode\n"), exit(1);
	if (st.st_size > 6)
		fprintf(stderr, "bad size\n"), exit(1);
}


static void test_mkdir(void)
{
	xsystem("mkdir /mnt/test/dir");
	xsystem("rmdir /mnt/test/dir");
}

static void check_mkdir(void)
{
	struct stat st;
	if (stat("/mnt/test/dir", &st) == -1) {
		if (errno == ENOENT)
			return;
		perror("stat"), exit(1);
	}
	if (!S_ISDIR(st.st_mode))
		fprintf(stderr, "bad mode\n"), exit(1);
}


#define N	512

static void test_create_more(void)
{
	int i;
	for (i = 0; i < N; i++) {
		char path[64];
		int h;
		snprintf(path, sizeof path, "/mnt/test/file-%d", i);
		h = creat(path, 0600);
		if (h == -1) perror(path), exit(1);
		if (close(h)) perror("close"), exit(1);
		sync();
	}
}

static void test_mkdir_more(void)
{
	int i;
	for (i = 0; i < N; i++) {
		char path[64];
		int h;
		snprintf(path, sizeof path, "/mnt/test/file-%d", i);
		h = mkdir(path, 0700);
		if (h == -1) perror(path), exit(1);
		sync();
	}
}

static void check_create_mkdir_more(void)
{
	int i;
	bool missing = false;
	for (i = 0; i < N; i++) {
		char path[64];
		int h;
		snprintf(path, sizeof path, "/mnt/test/file-%d", i);
		h = open(path, O_RDONLY);
		if (h == -1) {
			if (errno != ENOENT) perror(path), exit(1);
			missing = true;
		} else {
			if (missing) fprintf(stderr, "files not in order\n"), exit(1);
			if (close(h)) perror("close"), exit(1);
		}
	}
}


static void prepare_delete(void)
{
	xsystem("echo Hello > /mnt/test/a.txt");
}

static void test_delete(void)
{
	xsystem("rm /mnt/test/a.txt");
}


static void prepare_rename(void)
{
	xsystem("echo Hello > /mnt/test/a.txt");
}

static void test_rename(void)
{
	xsystem("mv /mnt/test/a.txt /mnt/test/b.txt");
}

static void check_rename(void)
{
	int h1, h2, h;
	h1 = open("/mnt/test/a.txt", O_RDONLY);
	h2 = open("/mnt/test/b.txt", O_RDONLY);
	if ((h1 < 0) + (h2 < 0) != 1) fprintf(stderr, "rename is not atomic\n"), exit(1);
	h = h1 >= 0 ? h1 : h2;
	if (close(h)) perror("close"), exit(1);
}


static void prepare_link(void)
{
	xsystem("echo Hello > /mnt/test/a.txt");
}

static void test_link(void)
{
	xsystem("ln /mnt/test/a.txt /mnt/test/b.txt");
}

static void check_link(void)
{
	struct stat st1, st2;
	unsigned long nlink;

	if (stat("/mnt/test/a.txt", &st1) == -1) {
		perror("stat"), exit(1);
	}
	if (!S_ISREG(st1.st_mode))
		fprintf(stderr, "bad mode\n"), exit(1);
	if (st1.st_size != 6)
		fprintf(stderr, "bad size\n"), exit(1);

	if (stat("/mnt/test/b.txt", &st2) == -1) {
		if (errno != ENOENT)
			perror("stat"), exit(1);
		nlink = 1;
	} else {
		if (!S_ISREG(st2.st_mode))
			fprintf(stderr, "bad mode\n"), exit(1);
		if (st2.st_size != 6)
			fprintf(stderr, "bad size\n"), exit(1);
		nlink = 2;
		if (st2.st_nlink != nlink)
			fprintf(stderr, "bad nlink\n"), exit(1);
	}

	if (st1.st_nlink != nlink)
		fprintf(stderr, "bad nlink\n"), exit(1);
}


static void test_symlink(void)
{
	xsystem("ln -s a.txt /mnt/test/b.txt");
}

static void check_symlink(void)
{
	char buf[1024];
	int r;
	r = readlink("/mnt/test/b.txt", buf, sizeof buf);
	if (r == -1) {
		if (errno == ENOENT)
			return;
		perror("readlink"), exit(1);
	}
	if (r == sizeof buf)
		fprintf(stderr, "readlink overflow\n"), exit(1);
	buf[r] = 0;
	if (!strcmp(buf, "a.txt"))
		return;
	fprintf(stderr, "bad symlink target: '%s'\n", buf);
	exit(1);
}


static void prepare_setfattr(void)
{
	xsystem("echo Hello > /mnt/test/a.txt");
}

static void test_setfattr(void)
{
	xsystem("setfattr -n user.us -v value /mnt/test/a.txt");
}


static void prepare_replaceattr(void)
{
	xsystem("echo Hello > /mnt/test/a.txt");
	xsystem("setfattr -n user.us -v value /mnt/test/a.txt");
}

static void test_replaceattr(void)
{
	xsystem("setfattr -n user.us -v value-value-2 /mnt/test/a.txt");
}


static void prepare_deleteattr(void)
{
	xsystem("echo Hello > /mnt/test/a.txt");
	xsystem("setfattr -n user.us -v value2 /mnt/test/a.txt");
}

static void test_deleteattr(void)
{
	xsystem("setfattr -x user.us /mnt/test/a.txt");
}


static const struct test tests[] = {
	{ "create", noop, test_create, check_create },
	{ "create_more", noop, test_create_more, check_create_mkdir_more },
	{ "mkdir", noop, test_mkdir, check_mkdir },
	{ "mkdir_more", noop, test_mkdir_more, check_create_mkdir_more },
	{ "delete", prepare_delete, test_delete, noop },
	{ "rename", prepare_rename, test_rename, check_rename },
	{ "link", prepare_link, test_link, check_link },
	{ "symlink", noop, test_symlink, check_symlink },
	{ "setfattr", prepare_setfattr, test_setfattr, noop },
	{ "replaceattr", prepare_replaceattr, test_replaceattr, noop },
	{ "deleteattr", prepare_deleteattr, test_deleteattr, noop },
};

int main(int argc, char *argv[])
{
	size_t t;
	const char *name = NULL;
	system("umount /mnt/test");
	if (argc >= 2) name = argv[1];
	for (t = 0; t < sizeof(tests) / sizeof(*tests); t++) {
		const struct test *test = &tests[t];
		if (name && strcasecmp(test->name, name))
			continue;
		do_test(test);
	}
	return 0;
}
